'''
Defines the main program classes for creating the GUI, drawing the blinking
selection rectangle, and maintaining information about the currently selected
accessible.

@author: Peter Parente
@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made 
available under the terms of the BSD which accompanies this distribution, and 
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
import pygtk
pygtk.require('2.0')
import gtk
import gobject
import gtk.glade
import gtk.gdk
import gettext, os, sys, locale
from icons import getIcon
import pyLinAcc
import os
import atk
import gnome
from accessible_treeview import *
from node import Node
from plugin_manager import *
from tools import Tools
from i18n import _
import wnck

GLADE_FILENAME = os.path.join(sys.prefix, 'share', 'accerciser', 'glade', 
                              'accerciser.glade')
if not os.path.exists(GLADE_FILENAME):
  GLADE_FILENAME = os.path.join(os.getcwd(), 'accerciser.glade')
  
class MainWindow(Tools):
  '''
  Class for the main accerciser window. 
  '''
  COL_ACC = 4
  COL_FILLED = 2
  
  def __init__(self):
    '''
    Gets references to important widgets, establishes event handlers,
    configures the tree view, and initializes the tree with the contents of the
    desktop.
    '''
    # mark the root of this window with its PID so we can easily identify it
    # as this app
    root_atk = atk.get_root()
    root_atk.set_description(str(os.getpid()))

    # parse the glade
    self.main_xml = gtk.glade.XML(GLADE_FILENAME, 'main_vbox', 
                                  gettext.textdomain())
    main_vbox = self.main_xml.get_widget('main_vbox')
    self.window = gtk.Window()
    self.window.set_title('accerciser')
    self.window.set_icon_name('accerciser')
    self.window.add(main_vbox)
    self.window.connect('destroy', self._onQuit)
    window_size = self.loadSettings('main_window_size')
    if window_size:
      width, height = window_size[0], window_size[1]
    else:
      width = height = 640
    print 'setting default size', width, height
    self.window.set_default_size(width, height)
    self.acc_treeview = AccessibleTreeView()
    self.node = self.acc_treeview.node
    self.node.connect('accessible_changed', self._onAccesibleChange)
    scrolled_window = self.main_xml.get_widget('scrolled_acc_tree')
    scrolled_window.add(self.acc_treeview)
    
    bin = self.main_xml.get_widget('alignment_topright')
    plugin_view1 = PluginView('Top right')
    bin.add(plugin_view1)
    bin = self.main_xml.get_widget('alignment_bottom')
    plugin_view2 = PluginView('Bottom panel')
    plugin_view2.connect('page_added', self._onBottomPanelChange, 'added')
    plugin_view2.connect('page_removed', self._onBottomPanelChange, 'removed')
    bin.add(plugin_view2)
    # load plugins
    self.plugin_manager = PluginManager(self.node, [plugin_view1, plugin_view2])
    self.plugin_manager.loadPlugins()

    # connect signal handlers and show the GUI in its initial state
    self.main_xml.signal_autoconnect(self)
    self.window.show_all()

    self.event_manager = pyLinAcc.Event.Manager()
    self.event_manager.addClient(self._accEventFocusChanged, 'focus')
    self.event_manager.addClient(self._accEventKeyPressed, 'keyboard:press')
    self.last_focused = None

  def _onRefreshAll(self, widget):
    '''
    Refreshes the entire tree at the desktop level.
    '''
    self.acc_treeview.refreshTopLevel()

  def _onRefreshCurrent(self, widget):
    self.acc_treeview.refreshCurrentLevel()

  def _onAccesibleChange(self, node, acc):
    # Update event monitor
    pass

  def _shutDown(self):
    self.saveSettings('main_window_size', 
                      (self.window.allocation.width,
                       self.window.allocation.height))
    self.event_manager.close()
    self.acc_treeview.destroy()
    self.plugin_manager.close()

  def _onQuit(self, widget):
    '''
    Quits the app.
    '''
    self._shutDown()
    gtk.main_quit()
    
  def _onAbout(self, widget):
    '''
    Shows the about dialog.
    '''
    xml = gtk.glade.XML(GLADE_FILENAME, 'about', gettext.textdomain())
    xml.signal_autoconnect(self)

  def _onAboutResponse(self, widget, response_id):
    '''
    close the about dialog.
    '''
    if response_id == gtk.RESPONSE_CANCEL:
      widget.destroy()
    
  def _onHelp(self, widget):
    '''
    Shows the help dialog.
    '''
    gnome.help_display('accerciser.xml')
    
  def _onCloseDialog(self, widget):
    '''
    Closes the window of the widget firing this event.
    '''
    widget.get_parent_window().destroy()

  def run(self):
    '''
    Runs the app.
    '''
    try:
      gtk.main()
    except KeyboardInterrupt:
      self._shutDown()

  def _onShowPlugins(self, widget):
     plugins_frame = self.main_xml.get_widget('frame_plugins')
     if widget.get_active():
        plugins_frame.show()
     else:
        plugins_frame.hide()
     
  def _onShowEventMonitor(self, widget):
    self.monitor_window.run()

  def _onSelectToggled(self, widget):
    if widget.get_active():
      self.event_manager.addClient(self._followFocus, 'focus')
    else:
      self.event_manager.removeClient(self._followFocus, 'focus')

  def _accEventFocusChanged(self, event):
    if not self.isMyApp(event.source):
      self.last_focused = event.source

  def _followFocus(self, event):
    if self.isMyApp(event.source):
      return
  
  def _accEventKeyPressed(self, event):
    if (event.any_data[1] & (1 << pyLinAcc.Constants.MODIFIER_META3) and
        event.any_data[0] == '?'):
      # Inspect last focused accessible
      if self.last_focused:
        self.node.update(self.last_focused)
    elif (event.any_data[1] & (1 << pyLinAcc.Constants.MODIFIER_META3) and
          event.any_data[0] == 'A'):
      # Inspect accessible under mouse
      display = gtk.gdk.Display(gtk.gdk.get_display())
      screen, x, y, flags =  display.get_pointer()
      desktop = pyLinAcc.Registry.getDesktop(0)
      wnck_screen = wnck.screen_get_default()
      window_order = [w.get_name() for w in wnck_screen.get_windows_stacked()]
      top_window = (None, -1)
      for app in desktop:
        if not app:
          continue
        for frame in app:
          if not frame:
            continue
          acc, z_order = self._getChildAccAtCoords(frame, x, y)
          if acc:
            try:
              z_order = window_order.index(frame.name)
            except ValueError:
              continue
            if z_order > top_window[1]:
              top_window = (acc, z_order)
      if top_window[0]:
        self.node.update(top_window[0])

  def _getChildAccAtCoords(self, parent, x, y):
    container = parent
    while container:
      try:
        ci = pyLinAcc.Interfaces.IComponent(container)
      except:
        return None, 0
      container =  ci.getAccessibleAtPoint(
        x, y, pyLinAcc.Constants.DESKTOP_COORDS)
      if container == pyLinAcc.Interfaces.IAccessible(ci):
        # The gecko bridge simply has getAccessibleAtPoint return itself
        # if there are no further children
        container = None
    if ci:
      acc = pyLinAcc.Interfaces.IAccessible(ci)
      if acc == parent:
        acc = None
      ci = pyLinAcc.Interfaces.IComponent(parent)
      z_order = ci.getMDIZOrder()
      return acc, z_order
    else:
      return None, 0

  def _onShowPlugins(self, widget):
    xml = gtk.glade.XML(GLADE_FILENAME, 
                        'dialog_plugins', 
                        gettext.textdomain())
    xml.signal_autoconnect(self)
    plugins_treeview = xml.get_widget('plugins_treeview')
    plugins_treeview.set_model(self.plugin_manager.plugins_store)
    crt = gtk.CellRendererText()
    tvc = gtk.TreeViewColumn('Name')
    tvc.pack_start(crt, True)
    tvc.set_attributes(crt, text=0)
    plugins_treeview.append_column(tvc)
    crc = gtk.CellRendererCombo()
    tvc = gtk.TreeViewColumn('View')
    tvc.pack_start(crc, False)
    tvc.set_expand(False)
    tvc.set_attributes(crc, text=2)
    crc.set_property('editable', True)
    crc.set_property('model', self.plugin_manager.views_store)
    crc.set_property('text-column', 1)
    crc.set_property('has-entry', True)
    crc.connect('edited', self._onViewChanged)
    plugins_treeview.append_column(tvc)
    d = xml.get_widget('dialog_plugins')
    d.show_all()

  def _onPluginsResponse(self, widget, response):
    widget.destroy()

  def _onViewChanged(self, cellrenderertext, path, new_text):
    self.plugin_manager.changeView(path, new_text)

  def _onBottomPanelChange(self, pluginview, page, page_num, action):
    paned = self.main_xml.get_widget('vpaned1')
    if not paned:
      return
    elif pluginview.get_n_pages() == 1 and action == 'added':
      paned.set_position(350)
    elif pluginview.get_n_pages() == 0:
      paned.set_position(paned.allocation.height - 30)
