'''
Defines the manager for plugin layout and loading.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2006, 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made 
available under the terms of the BSD which accompanies this distribution, and 
is available at U{http://www.opensource.org/licenses/bsd-license.php}
'''
import gtk
import gobject
import plugin as accerciser_plugin
from tools import Tools
import pyLinAcc
import os
import sys
import imp

PLUGIN_NOTEBOOK_GROUP = 1

class PluginView(gtk.Notebook):
  __gsignals__ = {'new_view' : 
                  (gobject.SIGNAL_RUN_FIRST,
                   gobject.TYPE_NONE, 
                   (gobject.TYPE_OBJECT,))}
  TARGET_PLUGINVIEW = 0
  TARGET_ROOTWIN = 1

  def __init__(self, view_name):
    gtk.Notebook.__init__(self)
    self.view_name = view_name
    self.set_scrollable(True)
    self.set_group_id(PLUGIN_NOTEBOOK_GROUP)
    self.connect('drag_end', self._onDragEnd)
    self.connect('drag_data_get', self._onDragDataGet)
    self.dest_type = None
    
  def _onDragDataGet(self, widget, context, selection_data, info, time):
    self.dest_type = info
    selection_data.set(selection_data.target, 8, '')

  def _onDragEnd(self, widget, drag_context):
    if self.dest_type == self.TARGET_PLUGINVIEW:
      return
    index = self.get_current_page()
    child = self.get_nth_page(index)
    self.emit('new_view', child)

  def append_page(self, child):
    name = child.child.plugin_name
    gtk.Notebook.append_page(self, child, gtk.Label(name))
    self.set_tab_detachable(child, True)
    self.set_tab_reorderable(child, True)
    
  def insert_page(self, child, position=-1):
    name = child.child.plugin_name
    gtk.Notebook.insert_page(self, child, gtk.Label(name), position)
    self.set_tab_detachable(child, True)
    self.set_tab_reorderable(child, True)

class PluginViewWindow(gtk.Window, Tools):
  def __init__(self, view_name):
    gtk.Window.__init__(self)
    self.plugin_view = PluginView(view_name)
    self.add(self.plugin_view)
    view_dimensions = self.loadSettings('plugin_view_sizes') or {}
    window_size = view_dimensions.get(view_name, (480, 480))
    self.set_default_size(window_size[0], window_size[1])
    self.set_title(view_name)
    self.set_position(gtk.WIN_POS_MOUSE)
    self.show_all()
  
class PluginManager(gobject.GObject, Tools):
  def __init__(self, node, pluginviews_main):
    gobject.GObject.__init__(self)
    self.node = node
    self.pluginviews_main = pluginviews_main
    self._restorePanedViews(pluginviews_main)
    self.views_store = gtk.ListStore(PluginView, str)
    for view in pluginviews_main:
      self.views_store.append([view, view.view_name])
    self.plugins_store = gtk.ListStore(str, str, str, object)
    self.closed = False

  def close(self):
    self._saveViewDimensions()
    for row in self.plugins_store:
      plugin = row[-1]
      plugin.close()
    self.closed = True

  def loadPlugins(self):
    plugin_layout = self._loadLayout()
    plugin_dir_local = os.path.join(os.environ['HOME'], '.accerciser', 'plugins')
    plugin_dir_global = os.path.join(sys.prefix, 'share', 'accerciser', 'plugins')
    plugin_list = []
    for plugin_dir in (plugin_dir_local, plugin_dir_global):
      if not os.path.isdir(plugin_dir):
        continue
      for fn in os.listdir(plugin_dir):
        if fn.endswith('.py') and not fn.startswith('.'):
          self._loadPlugin(plugin_dir, fn[:-3],
                           plugin_layout)
    for view, view_name in self.views_store:
      for child in view.get_children():
        plugin = child.child
        try:
          position = plugin_layout[view_name].index(plugin.plugin_name)
        except:
          position = -1
        view.reorder_child(child, position)
      self._connectSignals(view)
      view.show_all()
    self._buildPluginsModel()

  def _loadPlugin(self, plugin_dir, plugin_fn, plugin_layout):
    plugin_locals = {}
    sys.path.insert(0, plugin_dir)
    try:
      params = imp.find_module(plugin_fn, [plugin_dir])
      plugin = imp.load_module(plugin_fn, *params)
      plugin_locals = plugin.__dict__
    except SyntaxError, syntax_error:
      print 'Syntax error loading %s\n* %s' % (plugin_fn, str(syntax_error))
      return
    except Exception, e:
      print 'Error loading %s\n* %s' % (plugin_fn, str(e))
      return
    # use keys list to avoid size changes during iteration
    for symbol in plugin_locals.keys():
      try:
        is_plugin = issubclass(plugin_locals[symbol],
                               accerciser_plugin.Plugin)
      except TypeError:
        continue
      if is_plugin:
        # if a plugin class is found, initialize
        obj = plugin_locals[symbol](self.node)
        try:
          obj.init()
        except Exception, e:
          print 'Error initializing %s\n* %s' % (plugin_fn, str(e))
          return
        obj.acc = self.node.acc
        obj.onAccChanged(self.node.acc)
        pluginview = self._getViewForPlugin(obj.plugin_name, plugin_layout)
        if isinstance(obj, gtk.Widget):
          scrolled_w = gtk.ScrolledWindow()
          scrolled_w.set_policy(gtk.POLICY_AUTOMATIC, 
                                gtk.POLICY_AUTOMATIC)
          scrolled_w.set_border_width(3)
          scrolled_w.set_shadow_type(gtk.SHADOW_IN)
          scrolled_w.add(obj)
          pluginview.append_page(scrolled_w)
    sys.path.pop(0)

  def _getViewForPlugin(self, name, plugin_layout):
    for view_name in plugin_layout:
      if name in plugin_layout[view_name]:
        if view_name not in [row[1] for row in self.views_store]:
          w = PluginViewWindow(view_name)
          w.connect('delete_event', self._onPluginViewRemoved)
          self.views_store.append([w.plugin_view, view_name])
          return w.plugin_view
        else:
          for view, view_name2 in self.views_store:
            if view_name2 == view_name:
              return view
    return self.views_store[0][0]

  def _connectSignals(self, pluginview):
    pluginview.connect('new_view', self._onNewPluginView)
    pluginview.connect('page_added', self._onPluginLayoutChanged)
    pluginview.connect('page_removed', self._onPluginLayoutChanged)
    pluginview.connect('page_reordered', self._onPluginLayoutChanged)

  def _onNewPluginView(self, view, child):
    view.remove(child)
    self._newViewWithPage(child)

  def _newViewWithPage(self, page, view_name=None):
    if not view_name:
      view_name = 'Plugin View'
      view_num = 2
      while view_name in [row[1] for row in self.views_store]:
        view_name = 'Plugin View (%d)' % view_num
        view_num += 1
    w = PluginViewWindow(view_name)
    w.connect('delete_event', self._onPluginViewRemoved)
    pluginview = w.plugin_view
    self._connectSignals(pluginview)
    self.views_store.append([pluginview, view_name])
    pluginview.append_page(page)

  def _onPluginViewRemoved(self, pluginviewwindow, event):
    pluginview = pluginviewwindow.plugin_view
    for child in pluginview.get_children():
      pluginview.remove(child)
      self.views_store[0][0].append_page(child)

  def _onPluginLayoutChanged(self, pluginview, page, page_num):
    if self.closed:
      return
    iter = self.views_store.get_iter_root()
    while iter:
      view = self.views_store[iter][0]
      if view.get_n_pages() > 0 or view in self.pluginviews_main:
        iter = self.views_store.iter_next(iter)
      else:
        if view.get_n_pages() == 0:
          self._onPluginViewRemoved(view.parent, None)
          view.parent.destroy()
        if not self.views_store.remove(iter): break
    iter = self.plugins_store.get_iter_root()
    while iter:
      view = self.plugins_store[iter][-1].parent.parent
      if view:
        self.plugins_store[iter][2] = view.view_name
      iter = self.plugins_store.iter_next(iter)
    self._saveLayout()

  def _buildPluginsModel(self):
    for view, view_name in self.views_store:
      for child in view.get_children():
        plugin = child.child
        self.plugins_store.append([plugin.plugin_name,
                                   plugin.plugin_description,
                                   view_name, plugin])
    
  def _loadLayout(self):
    rv = self.loadSettings('plugin_layout') or {}
    return rv

  def _saveLayout(self):
    layout = {}
    for view, view_name in self.views_store:
      layout[view_name] = []
      for child in view.get_children():
        layout[view_name].append(child.child.plugin_name)
    self.saveSettings('plugin_layout', layout)

  def _saveViewDimensions(self):
    view_dimensions = {}
    for view, view_name in self.views_store:
      child = view
      while child:
        if isinstance(child.parent, PluginViewWindow):
          window = child.parent
          view_dimensions[view_name] = (window.allocation.width, 
                                        window.allocation.height)
          break
        elif isinstance(child.parent, gtk.Paned):
          paned = child.parent
          view_dimensions[view_name] = (paned.get_position(),)
          break
        child = child.parent
    self.saveSettings('plugin_view_sizes', view_dimensions)
  
  def _restorePanedViews(self, views):
    view_dimensions = self.loadSettings('plugin_view_sizes') or {}
    for view in views:
      if not view_dimensions.has_key(view.view_name):
        continue
      child = view
      while child:
        if isinstance(child.parent, gtk.Paned):
          paned = child.parent
          paned.set_position(view_dimensions[view.view_name][0])
          break
        child = child.parent
    self.saveSettings('plugin_view_sizes', view_dimensions)

  def changeView(self, path, new_text):
    plugin = self.plugins_store[path][-1]
    old_view = plugin.parent.parent
    view_names = [v[1] for v in self.views_store]
    try:
      view = self.views_store[view_names.index(new_text)][0]
    except ValueError:
      old_view.remove(plugin.parent)
      self._newViewWithPage(plugin.parent, new_text)
      return

    if plugin.parent in view.get_children():
      return
    else:
      old_view.remove(plugin.parent)
      view.append_page(plugin.parent)
      
